@@ -26,12 +26,13 @@
 
 #include <stdint.h>
 #include <errno.h>
+#include <stdbool.h>
 #include <string.h>
 #include <sys/socket.h>
 
 #include <linux/socket.h>
-#include <linux/nfc.h>
 
+#include <near/nfc_copy.h>
 #include <near/plugin.h>
 #include <near/log.h>
 #include <near/types.h>
@@ -74,7 +75,7 @@
 #define T4_ALL_ACCESS		0x00
 #define T4_READ_ONLY		0xFF
 
-#define APDU_STATUS(a) (g_ntohs(*((uint16_t *)(a))))
+#define APDU_STATUS(a) near_get_be16(a)
 
 /* Tag Type 4 version ID */
 static uint8_t iso_appname_v1[] = { 0xd2, 0x76, 0x0, 0x0, 0x85, 0x01, 0x0 };
@@ -164,7 +165,7 @@ static int t4_cookie_release(int err, void *data)
 
 	DBG("%p", cookie);
 
-	if (cookie == NULL)
+	if (!cookie)
 		return err;
 
 	if (err < 0 && cookie->cb)
@@ -186,7 +187,7 @@ static int ISO_send_cmd(uint8_t class,
 			uint8_t param2,
 			uint8_t *cmd_data,
 			uint8_t cmd_data_length,
-			near_bool_t le,
+			bool le,
 			near_recv cb,
 			void *in_data)
 {
@@ -198,7 +199,7 @@ static int ISO_send_cmd(uint8_t class,
 	DBG("CLA-%02x INS-%02x P1-%02x P2-%02x",
 			class, instruction, param1, param2);
 
-	if (le == FALSE) {
+	if (!le) {
 		if (cmd_data)
 			total_cmd_length = APDU_HEADER_LEN + cmd_data_length;
 		else
@@ -208,7 +209,7 @@ static int ISO_send_cmd(uint8_t class,
 	}
 
 	cmd = g_try_malloc0(total_cmd_length);
-	if (cmd == NULL) {
+	if (!cmd) {
 		DBG("Mem alloc failed");
 		err = -ENOMEM;
 		goto out_err;
@@ -224,7 +225,7 @@ static int ISO_send_cmd(uint8_t class,
 		memcpy(cmd->data, cmd_data, cmd_data_length);
 		/* The Le byte set to 0x00 defined that any length
 		 * of PICC response is allowed */
-		if (le == TRUE)
+		if (le)
 			cmd->data[cmd_data_length] = 0;
 	}
 
@@ -255,7 +256,7 @@ static int ISO_Select(uint8_t *filename, uint8_t fnamelen, uint8_t P1,
 			0x00,		/* P2: First or only occurrence */
 			filename,	/* cmd_data */
 			fnamelen,	/* uint8_t cmd_data_length*/
-			FALSE,
+			false,
 			cb,
 			cookie);
 }
@@ -272,7 +273,7 @@ static int ISO_ReadBinary(uint16_t offset, uint8_t readsize,
 			(uint8_t) (offset & 0xFF),
 			0,		/* no data send */
 			readsize,	/* bytes to read */
-			FALSE,
+			false,
 			cb,
 			cookie);
 }
@@ -289,7 +290,7 @@ static int ISO_Update(uint16_t offset, uint8_t nlen,
 			(uint8_t) (offset & 0xFF),
 			data,			/* length of NDEF data */
 			nlen,			/* NLEN + NDEF data */
-			FALSE,
+			false,
 			cb,
 			cookie);
 }
@@ -329,7 +330,7 @@ static int data_read_cb(uint8_t *resp, int length, void *data)
 
 		DBG("Done reading");
 
-		records = near_ndef_parse(nfc_data, data_length);
+		records = near_ndef_parse_msg(nfc_data, data_length, NULL);
 		near_tag_add_records(cookie->tag, records, cookie->cb, 0);
 
 		return t4_cookie_release(0, cookie);
@@ -365,12 +366,12 @@ static int t4_readbin_NDEF_ID(uint8_t *resp, int length, void *data)
 
 	/* Add data to the tag */
 	err = near_tag_add_data(cookie->adapter_idx, cookie->target_idx, NULL,
-				g_ntohs(*((uint16_t *)(resp + NFC_STATUS_BYTE_LEN))));
+				near_get_be16(resp + NFC_STATUS_BYTE_LEN));
 	if (err < 0)
 		return t4_cookie_release(err, cookie);
 
 	tag = near_tag_get_tag(cookie->adapter_idx, cookie->target_idx);
-	if (tag == NULL)
+	if (!tag)
 		return t4_cookie_release(-ENOMEM, cookie);
 
 	near_tag_set_max_ndef_size(tag, cookie->max_ndef_size);
@@ -420,7 +421,7 @@ static int t4_select_NDEF_ID(uint8_t *resp, int length, void *data)
 static int t4_readbin_cc(uint8_t *resp, int length, void *data)
 {
 	struct t4_cookie *cookie = data;
-	struct type4_cc *read_cc;
+	struct type4_cc *read_cc = (struct type4_cc *)&resp[1];
 
 	DBG("%d", length);
 
@@ -434,16 +435,6 @@ static int t4_readbin_cc(uint8_t *resp, int length, void *data)
 		return t4_cookie_release(-EIO, cookie);
 	}
 
-	/* -2 for status word and -1 is for NFC first byte... */
-	read_cc = g_try_malloc0(length - 2 - NFC_STATUS_BYTE_LEN);
-	if (read_cc == NULL) {
-		DBG("Mem alloc failed");
-
-		return t4_cookie_release(-ENOMEM, cookie);
-	}
-
-	memcpy(read_cc, &resp[1], length - 2 - NFC_STATUS_BYTE_LEN) ;
-
 	cookie->r_apdu_max_size = g_ntohs(read_cc->max_R_apdu_data_size) -
 			APDU_HEADER_LEN;
 	cookie->c_apdu_max_size = g_ntohs(read_cc->max_C_apdu_data_size);
@@ -486,7 +477,7 @@ static int t4_select_cc(uint8_t *resp, int length, void *data)
 
 		cookie->tag = near_tag_get_tag(cookie->adapter_idx,
 						cookie->target_idx);
-		if (cookie->tag == NULL)
+		if (!cookie->tag)
 			return t4_cookie_release(-ENOMEM, cookie);
 
 		near_tag_set_blank(cookie->tag, TRUE);
@@ -531,7 +522,7 @@ static int t4_select_file_by_name_v2(uint8_t *resp, int length, void *data)
 	DBG("%d", length);
 
 	if (length < 0)
-		t4_cookie_release(length, cookie);
+		return t4_cookie_release(length, cookie);
 
 	/* Check for APDU error - Not found */
 	if (APDU_STATUS(resp + STATUS_WORD_1) == APDU_NOT_FOUND) {
@@ -563,7 +554,7 @@ static int nfctype4_read(uint32_t adapter_idx,
 	DBG("");
 
 	cookie = g_try_malloc0(sizeof(struct t4_cookie));
-	if (cookie == NULL)
+	if (!cookie)
 		return -ENOMEM;
 
 	cookie->adapter_idx = adapter_idx;
@@ -634,10 +625,10 @@ static int data_write(uint32_t adapter_idx, uint32_t target_idx,
 
 	cookie = g_try_malloc0(sizeof(struct t4_cookie));
 
-	if (cookie == NULL) {
+	if (!cookie) {
 		err = -ENOMEM;
 
-		if (cb != NULL)
+		if (cb)
 			cb(adapter_idx, target_idx, err);
 
 		return err;
@@ -690,13 +681,13 @@ static int nfctype4_write(uint32_t adapter_idx, uint32_t target_idx,
 
 	DBG("");
 
-	if (ndef == NULL || cb == NULL) {
+	if (!ndef || !cb) {
 		err = -EINVAL;
 		goto out_err;
 	}
 
 	tag = near_tag_get_tag(adapter_idx, target_idx);
-	if (tag == NULL) {
+	if (!tag) {
 		err = -EINVAL;
 		goto out_err;
 	}
@@ -704,7 +695,7 @@ static int nfctype4_write(uint32_t adapter_idx, uint32_t target_idx,
 	err = data_write(adapter_idx, target_idx, ndef, tag, cb);
 
 out_err:
-	if (cb != NULL)
+	if (cb)
 		cb(adapter_idx, target_idx, err);
 
 	return err;
@@ -735,7 +726,7 @@ static int nfctype4_check_presence(uint32_t adapter_idx,
 	DBG("");
 
 	cookie = g_try_malloc0(sizeof(struct t4_cookie));
-	if (cookie == NULL)
+	if (!cookie)
 		return -ENOMEM;
 
 	cookie->adapter_idx = adapter_idx;
@@ -797,7 +788,7 @@ static int read_cc_file(uint8_t *resp, int length, void *data)
 
 	/* -2 for status word and -1 is for NFC first byte... */
 	read_cc = g_try_malloc0(length - 2 - NFC_STATUS_BYTE_LEN);
-	if (read_cc == NULL) {
+	if (!read_cc) {
 		err = -ENOMEM;
 		goto out_err;
 	}
@@ -807,7 +798,7 @@ static int read_cc_file(uint8_t *resp, int length, void *data)
 	cookie->max_ndef_size = g_ntohs(read_cc->tlv_fc.max_ndef_size);
 
 	tag = near_tag_get_tag(cookie->adapter_idx, cookie->target_idx);
-	if (tag == NULL) {
+	if (!tag) {
 		err = -EINVAL;
 		goto out_err;
 	}
@@ -919,7 +910,7 @@ static int format_resp(uint8_t *resp, int length, void *data)
 	}
 
 	tag = near_tag_get_tag(cookie->adapter_idx, cookie->target_idx);
-	if (tag == NULL)
+	if (!tag)
 		return t4_cookie_release(-EINVAL, cookie);
 
 	DBG("Formatting is done");
@@ -977,7 +968,7 @@ static int write_data_to_ndef_file(uint8_t *resp, int length, void *data)
 				+ ARRAY_SIZE(ndef_nlen);
 
 	cmd_data = g_try_malloc0(cmd_data_length);
-	if (cmd_data == NULL) {
+	if (!cmd_data) {
 		err = -ENOMEM;
 		goto out_err;
 	}
@@ -990,7 +981,7 @@ static int write_data_to_ndef_file(uint8_t *resp, int length, void *data)
 
 	err = ISO_send_cmd(PICC_CLASS, WRITE_DATA_TO_FILE,
 				0x00, 0x00, cmd_data, cmd_data_length,
-				TRUE, format_resp, cookie);
+				true, format_resp, cookie);
 	if (err < 0) {
 		near_error("wrtie data to ndef file req failed %d", err);
 		goto out_err;
@@ -1028,7 +1019,7 @@ static int create_ndef_file(uint8_t *resp, int length, void *data)
 	}
 
 	ndef = g_try_malloc0(sizeof(struct desfire_std_file));
-	if (ndef == NULL) {
+	if (!ndef) {
 		err = -ENOMEM;
 		goto out_err;
 	}
@@ -1045,7 +1036,7 @@ static int create_ndef_file(uint8_t *resp, int length, void *data)
 	err = ISO_send_cmd(PICC_CLASS, CREATE_STD_DATA_FILE,
 				0x00, 0x00, (uint8_t *)ndef,
 				sizeof(struct desfire_std_file),
-				TRUE, write_data_to_ndef_file, cookie);
+				true, write_data_to_ndef_file, cookie);
 	if (err < 0) {
 		near_error("create ndef file req failed %d", err);
 		goto out_err;
@@ -1089,7 +1080,7 @@ static int write_data_to_cc_file(uint8_t *resp, int length, void *data)
 	}
 
 	cc = g_try_malloc0(sizeof(struct desfire_cc_file));
-	if (cc == NULL) {
+	if (!cc) {
 		err = -ENOMEM;
 		goto out_err;
 	}
@@ -1110,7 +1101,7 @@ static int write_data_to_cc_file(uint8_t *resp, int length, void *data)
 	err = ISO_send_cmd(PICC_CLASS, WRITE_DATA_TO_FILE,
 				0x00, 0x00, (uint8_t *)cc,
 				sizeof(struct desfire_cc_file),
-				TRUE, create_ndef_file, cookie);
+				true, create_ndef_file, cookie);
 	if (err < 0) {
 		near_error("write data to cc file req failed %d", err);
 		goto out_err;
@@ -1149,7 +1140,7 @@ static int create_cc_file(uint8_t *resp, int length, void *data)
 	}
 
 	cc = g_try_malloc0(sizeof(struct desfire_std_file));
-	if (cc == NULL) {
+	if (!cc) {
 		err = -ENOMEM;
 		goto out_err;
 	}
@@ -1165,7 +1156,7 @@ static int create_cc_file(uint8_t *resp, int length, void *data)
 				CREATE_STD_DATA_FILE,
 				0x00, 0x00, (uint8_t *)cc,
 				sizeof(struct desfire_std_file),
-				TRUE, write_data_to_cc_file, cookie);
+				true, write_data_to_cc_file, cookie);
 	if (err < 0) {
 		near_error("create cc file req failed %d", err);
 		goto out_err;
@@ -1205,7 +1196,7 @@ static int select_application_1(uint8_t *resp, int length, void *data)
 	/* Step4 : Select application (which is created just now) */
 	cmd_data_length = ARRAY_SIZE(desfire_aid_1);
 	cmd_data = g_try_malloc0(cmd_data_length);
-	if (cmd_data == NULL) {
+	if (!cmd_data) {
 		err = -ENOMEM;
 		goto out_err;
 	}
@@ -1213,7 +1204,7 @@ static int select_application_1(uint8_t *resp, int length, void *data)
 	memcpy(cmd_data, desfire_aid_1, cmd_data_length);
 	err = ISO_send_cmd(PICC_CLASS, SELECT_APPLICATION,
 				0x00, 0x00, cmd_data, cmd_data_length,
-				TRUE, create_cc_file, cookie);
+				true, create_cc_file, cookie);
 	if (err < 0) {
 		near_error("select application1 req failed %d", err);
 		goto out_err;
@@ -1251,7 +1242,7 @@ static int create_application(uint8_t *resp, int length, void *data)
 	}
 
 	app = g_try_malloc0(sizeof(struct desfire_app));
-	if (app == NULL) {
+	if (!app) {
 		err = -ENOMEM;
 		goto out_err;
 	}
@@ -1266,7 +1257,7 @@ static int create_application(uint8_t *resp, int length, void *data)
 	err = ISO_send_cmd(PICC_CLASS, CREATE_APPLICATION,
 				0x00, 0x00, (uint8_t *)app,
 				sizeof(struct desfire_app),
-				TRUE, select_application_1, cookie);
+				true, select_application_1, cookie);
 	if (err < 0) {
 		near_error("create application req failed %d", err);
 		goto out_err;
@@ -1306,7 +1297,7 @@ static int select_application(uint8_t *resp, int length, void *data)
 	/* AID : 000000h */
 	cmd_data_length = ARRAY_SIZE(desfire_aid);
 	cmd_data = g_try_malloc0(cmd_data_length);
-	if (cmd_data == NULL) {
+	if (!cmd_data) {
 		err = -ENOMEM;
 		goto out_err;
 	}
@@ -1316,7 +1307,7 @@ static int select_application(uint8_t *resp, int length, void *data)
 	err = ISO_send_cmd(PICC_CLASS,
 				SELECT_APPLICATION,
 				0x00, 0x00, cmd_data, cmd_data_length,
-				TRUE, create_application, cookie);
+				true, create_application, cookie);
 	if (err < 0) {
 		near_error("select application req failed %d", err);
 		goto out_err;
@@ -1348,7 +1339,7 @@ static int get_version_frame3(uint8_t *resp, int length, void *data)
 
 		err = ISO_send_cmd(PICC_CLASS,
 					GET_VERSION_FRAME_RESPONSE_BYTE,
-					0x00, 0x00, NULL, 0, FALSE,
+					0x00, 0x00, NULL, 0, false,
 					select_application, cookie);
 		if (err < 0) {
 			near_error("get version3 req failed %d", err);
@@ -1387,7 +1378,7 @@ static int get_version_frame2(uint8_t *resp, int length, void *data)
 		cookie->memory_size = (1 << (resp[6] / 2));
 		err = ISO_send_cmd(PICC_CLASS,
 					GET_VERSION_FRAME_RESPONSE_BYTE,
-					0x00, 0x00, NULL, 0, FALSE,
+					0x00, 0x00, NULL, 0, false,
 					get_version_frame3, cookie);
 		if (err < 0) {
 			near_error("get version2 req failed %d", err);
@@ -1434,7 +1425,7 @@ static int nfctype4_format(uint32_t adapter_idx, uint32_t target_idx,
 	DBG("");
 
 	cookie = g_try_malloc0(sizeof(struct t4_cookie));
-	if (cookie == NULL)
+	if (!cookie)
 		return -ENOMEM;
 
 	cookie->adapter_idx = adapter_idx;
@@ -1443,7 +1434,7 @@ static int nfctype4_format(uint32_t adapter_idx, uint32_t target_idx,
 
 	/* Step1 : Get Version */
 	err = ISO_send_cmd(PICC_CLASS, GET_VERSION,
-				0x00, 0x00, NULL, 0, FALSE,
+				0x00, 0x00, NULL, 0, false,
 				get_version_frame2, cookie);
 
 	if (err < 0) {
@@ -1454,7 +1445,7 @@ static int nfctype4_format(uint32_t adapter_idx, uint32_t target_idx,
 	return err;
 }
 
-static struct near_tag_driver type4_driver = {
+static struct near_tag_driver type4a_driver = {
 	.type           = NFC_PROTO_ISO14443,
 	.priority       = NEAR_TAG_PRIORITY_DEFAULT,
 	.read           = nfctype4_read,
@@ -1463,18 +1454,34 @@ static struct near_tag_driver type4_driver = {
 	.format		= nfctype4_format,
 };
 
+static struct near_tag_driver type4b_driver = {
+	.type           = NFC_PROTO_ISO14443_B,
+	.priority       = NEAR_TAG_PRIORITY_DEFAULT,
+	.read           = nfctype4_read,
+	.write          = nfctype4_write,
+	.check_presence = nfctype4_check_presence,
+	.format		= nfctype4_format,
+};
+
 static int nfctype4_init(void)
 {
+	int ret;
+
 	DBG("");
 
-	return near_tag_driver_register(&type4_driver);
+	ret = near_tag_driver_register(&type4b_driver);
+	if (ret < 0)
+		return ret;
+
+	return near_tag_driver_register(&type4a_driver);
 }
 
 static void nfctype4_exit(void)
 {
 	DBG("");
 
-	near_tag_driver_unregister(&type4_driver);
+	near_tag_driver_unregister(&type4a_driver);
+	near_tag_driver_unregister(&type4b_driver);
 }
 
 NEAR_PLUGIN_DEFINE(nfctype4, "NFC Forum Type 4 tags support", VERSION,
