@@ -61,17 +61,11 @@ static GSList *driver_list = NULL;
 static void free_device(gpointer data)
 {
 	struct near_device *device = data;
-	GList *list;
 
 	DBG("device %p", device);
 
-	for (list = device->records; list; list = list->next) {
-		struct near_ndef_record *record = list->data;
+	near_ndef_records_free(device->records);
 
-		__near_ndef_record_free(record);
-	}
-
-	g_list_free(device->records);
 	g_free(device->path);
 	g_free(device->data);
 	g_free(device);
@@ -87,7 +81,7 @@ struct near_device *near_device_get_device(uint32_t adapter_idx,
 
 	path = g_strdup_printf("%s/nfc%d/device%d", NFC_PATH,
 					adapter_idx, target_idx);
-	if (path == NULL)
+	if (!path)
 		return NULL;
 
 	device = g_hash_table_lookup(device_hash, path);
@@ -97,21 +91,6 @@ struct near_device *near_device_get_device(uint32_t adapter_idx,
 	return device;
 }
 
-void __near_device_remove(struct near_device *device)
-{
-	char *path = device->path;
-
-	DBG("path %s", device->path);
-
-	if (g_hash_table_lookup(device_hash, device->path) == NULL)
-		return;
-
-	g_dbus_unregister_interface(connection, device->path,
-						NFC_DEVICE_INTERFACE);
-
-	g_hash_table_remove(device_hash, path);
-}
-
 const char *__near_device_get_path(struct near_device *device)
 {
 	return device->path;
@@ -134,7 +113,7 @@ static void append_records(DBusMessageIter *iter, void *user_data)
 		char *path;
 
 		path = __near_ndef_record_get_path(record);
-		if (path == NULL)
+		if (!path)
 			continue;
 
 		dbus_message_iter_append_basic(iter, DBUS_TYPE_OBJECT_PATH,
@@ -152,7 +131,7 @@ static DBusMessage *get_properties(DBusConnection *conn,
 	DBG("conn %p", conn);
 
 	reply = dbus_message_new_method_return(msg);
-	if (reply == NULL)
+	if (!reply)
 		return NULL;
 
 	dbus_message_iter_init_append(reply, &array);
@@ -186,12 +165,12 @@ static void push_cb(uint32_t adapter_idx, uint32_t target_idx, int status)
 	conn = near_dbus_get_connection();
 	device = near_device_get_device(adapter_idx, target_idx);
 
-	if (conn == NULL || device == NULL)
+	if (!conn || !device)
 		return;
 
 	if (status != 0) {
-		reply = __near_error_failed(device->push_msg, EINVAL);
-		if (reply != NULL)
+		reply = __near_error_failed(device->push_msg, -status);
+		if (reply)
 			g_dbus_send_message(conn, reply);
 	} else {
 		g_dbus_send_reply(conn, device->push_msg, DBUS_TYPE_INVALID);
@@ -240,6 +219,10 @@ static char *sn_from_message(DBusMessage *msg)
 				return NEAR_DEVICE_SN_SNEP;
 			else if (g_strcmp0(value, "Handover") == 0)
 				return NEAR_DEVICE_SN_HANDOVER;
+			else if (g_strcmp0(value, "StaticHandover") == 0)
+				return NEAR_DEVICE_SN_HANDOVER;
+			else if (g_strcmp0(value, "MIME") == 0)
+				return NEAR_DEVICE_SN_SNEP;
 			else
 				return NULL;
 
@@ -268,13 +251,13 @@ static DBusMessage *push_ndef(DBusConnection *conn,
 	device->push_msg = dbus_message_ref(msg);
 
 	service_name = sn_from_message(msg);
-	if (service_name == NULL) {
+	if (!service_name) {
 		err = -EINVAL;
 		goto error;
 	}
 
 	ndef = __ndef_build_from_message(msg);
-	if (ndef == NULL) {
+	if (!ndef) {
 		err = -EINVAL;
 		goto error;
 	}
@@ -283,9 +266,6 @@ static DBusMessage *push_ndef(DBusConnection *conn,
 	if (err < 0)
 		goto error;
 
-	g_free(ndef);
-	g_free(ndef->data);
-
 	return NULL;
 
 error:
@@ -313,21 +293,39 @@ static const GDBusSignalTable device_signals[] = {
 	{ }
 };
 
+void __near_device_remove(struct near_device *device)
+{
+	char *path = device->path;
+
+	DBG("path %s", device->path);
+
+	if (!g_hash_table_lookup(device_hash, device->path))
+		return;
+
+	if (device->push_msg)
+		push_cb(device->adapter_idx, device->target_idx, EIO);
+
+	g_dbus_unregister_interface(connection, device->path,
+						NFC_DEVICE_INTERFACE);
+
+	g_hash_table_remove(device_hash, path);
+}
+
 int near_device_add_data(uint32_t adapter_idx, uint32_t target_idx,
 			uint8_t *data, size_t data_length)
 {
 	struct near_device *device;
 
 	device = near_device_get_device(adapter_idx, target_idx);
-	if (device == NULL)
+	if (!device)
 		return -ENODEV;
 
 	device->data_length = data_length;
 	device->data = g_try_malloc0(data_length);
-	if (device->data == NULL)
+	if (!device->data)
 		return -ENOMEM;
 
-	if (data != NULL)
+	if (data)
 		memcpy(device->data, data, data_length);
 
 	return 0;
@@ -342,6 +340,9 @@ int near_device_add_records(struct near_device *device, GList *records,
 
 	DBG("records %p", records);
 
+	near_ndef_records_free(device->records);
+	device->records = NULL;
+
 	for (list = records; list; list = list->next) {
 		record = list->data;
 
@@ -349,7 +350,7 @@ int near_device_add_records(struct near_device *device, GList *records,
 					NFC_PATH, device->adapter_idx,
 					device->target_idx, device->n_records);
 
-		if (path == NULL)
+		if (!path)
 			continue;
 
 		__near_ndef_record_register(record, path);
@@ -360,7 +361,12 @@ int near_device_add_records(struct near_device *device, GList *records,
 
 	__near_agent_ndef_parse_records(device->records);
 
-	if (cb != NULL)
+	near_dbus_property_changed_array(device->path,
+					NFC_DEVICE_INTERFACE, "Records",
+					DBUS_TYPE_OBJECT_PATH, append_records,
+					device);
+
+	if (cb)
 		cb(device->adapter_idx, device->target_idx, status);
 
 	g_list_free(records);
@@ -375,16 +381,16 @@ struct near_device *__near_device_add(uint32_t adapter_idx, uint32_t target_idx,
 	char *path;
 
 	device = near_device_get_device(adapter_idx, target_idx);
-	if (device != NULL)
+	if (device)
 		return NULL;
 
 	device = g_try_malloc0(sizeof(struct near_device));
-	if (device == NULL)
+	if (!device)
 		return NULL;
 
 	device->path = g_strdup_printf("%s/nfc%d/device%d", NFC_PATH,
 					adapter_idx, target_idx);
-	if (device->path == NULL) {
+	if (!device->path) {
 		g_free(device);
 		return NULL;
 	}
@@ -398,7 +404,7 @@ struct near_device *__near_device_add(uint32_t adapter_idx, uint32_t target_idx,
 	}
 
 	path = g_strdup(device->path);
-	if (path == NULL) {
+	if (!path) {
 		g_free(device);
 		return NULL;
 	}
@@ -438,7 +444,7 @@ int __near_device_push(struct near_device *device,
 
 	DBG("");
 
-	if (__near_adapter_get_dep_state(device->adapter_idx) == FALSE) {
+	if (!__near_adapter_get_dep_state(device->adapter_idx)) {
 		near_error("DEP link is not established");
 		return -ENOLINK;
 	}
@@ -465,7 +471,7 @@ int near_device_driver_register(struct near_device_driver *driver)
 {
 	DBG("");
 
-	if (driver->listen == NULL)
+	if (!driver->listen)
 		return -EINVAL;
 
 	driver_list = g_slist_insert_sorted(driver_list, driver, cmp_prio);
